/* 
TODO:
- POST
- HTTP Errors
- HTTP header variables
*/

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <netinet/in.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>


#include "WebForm.h"

static int
sockprintf (int socket, char *format,...)
{
  va_list ap;
  char buf[10240];

  va_start (ap, format);
  vsprintf (buf, format, ap);
  va_end (ap);
  return write (socket, buf, strlen (buf));
}

WebForm::WebForm()
{
  HTMLPage = NULL;
  HTMLPageLen = 0;
  HTTPError = 0;
  URL = NULL;
  Host = NULL;
  Referer = NULL;
  Cookie = NULL;
  UserAgent = NULL;
  URLParams = NULL;
  POSTParams = NULL;
  Debug = 0;
}

WebForm::~WebForm()
{
  Reset();
}

void WebForm::DebugMode(int on)
{
  Debug = on;
}

void WebForm::Reset(void)
{
  struct WebFormParam *Param, *Next;

  if (HTMLPage)
    free(HTMLPage);
  if (URL)
    free(URL);
  if (Host)
    free(Host);
  if (Referer)
    free(Referer);
  if (Cookie)
    free(Cookie);
  if (UserAgent)
    free(UserAgent);

  Param = URLParams;
  while (Param) {
   Next = Param->Next;
   free(Param->Value);
   free(Param->Name);
   free(Param);
   Param = Next;
  }

  Param = POSTParams;
  while (Param) {
   Next = Param->Next;
   free(Param->Value);
   free(Param->Name);
   free(Param);
   Param = Next;
  }

  HTMLPage = NULL;
  HTMLPageLen = 0;
  HTTPError = 0;
  URL = NULL;
  Host = NULL;
  Port = 0;
  Referer = NULL;
  Cookie = NULL;
  UserAgent = NULL;
  URLParams = NULL;
  POSTParams = NULL;
}

int WebForm::GetHTTPError(void)
{
  return HTTPError;
}

int WebForm::SetURL(char *URL)
{
  if (URL == NULL)
    return -1;

  if (this->URL)
   free(this->URL);
  this->URL = strdup(URL);
  if (!this->URL)
   return -1;
  return 0;
}

int WebForm::SetReferer(char *Referer)
{
  if (Referer == NULL)
    return -1;

  if (this->Referer)
   free(this->Referer);
  this->Referer = strdup(Referer);
  if (!this->Referer)
   return -1;
  return 0;
}

int WebForm::SetCookie(char *Cookie)
{
  if (Cookie == NULL)
    return -1;

  if (this->Cookie)
   free(this->Cookie);
  this->Cookie = strdup(Cookie);
  if (!this->Cookie)
   return -1;
  return 0;
}

int WebForm::SetUserAgent(char *UserAgent)
{
  if (UserAgent == NULL)
    return -1;

  if (this->UserAgent)
   free(this->UserAgent);
  this->UserAgent = strdup(UserAgent);
  if (!this->UserAgent)
   return -1;
  return 0;
}


int WebForm::SetHost(char *Host, int Port)
{
  if (Host == NULL)
    return -1;
  if (Port < 1 || Port > 65535)
    return -1;

  if (this->Host)
   free(this->Host);
  this->Host = strdup(Host);
  if (!this->Host)
   return -1;
  this->Port = Port;
  return 0;
}

int WebForm::SetURLParam(char *Name, char *Value)
{
  unsigned char *nValue;
  struct WebFormParam *Param;

  if (Name == NULL)
    return -1;

  (char *)nValue = Value!=NULL?strdup((char*)Value):strdup("");
  if (!nValue)
    return -1;

  Param = URLParams;
  while (Param) {
    if (!strcasecmp((char*)Param->Name, Name)) {
      free(Param->Value);
      Param->Value = (unsigned char *)nValue;
      return 0;
    }
    Param = Param->Next;
  }

  Param = (struct WebFormParam*)malloc(sizeof(struct WebFormParam));
  if (Param == NULL) {
    free(nValue);
    return -1;
  }
  Param->Name = strdup(Name);
  if (Param->Name == NULL) {
    free(nValue);
    free(Param);
    return -1;
  }
  Param->Value = nValue;
  Param->Next = URLParams;
  URLParams = Param;
  return 0;
}

int WebForm::SetPOSTParam(char *Name, char *Value)
{
  unsigned char *nValue;
  struct WebFormParam *Param;

  if (Name == NULL)
    return -1;

  (char *)nValue = Value!=NULL?strdup(Value):strdup("");
  if (!nValue)
    return -1;

  Param = POSTParams;
  while (Param) {
    if (!strcasecmp(Param->Name, Name)) {
      free(Param->Value);
      Param->Value = nValue;
      return 0;
    }
    Param = Param->Next;
  }

  Param = (struct WebFormParam*)malloc(sizeof(struct WebFormParam));
  if (Param == NULL) {
    free(nValue);
    return -1;
  }
  Param->Name = strdup(Name);
  if (Param->Name == NULL) {
    free(nValue);
    free(Param);
    return -1;
  }
  Param->Value = nValue;
  Param->Next = POSTParams;
  POSTParams = Param;
  return 0;
}

char *WebForm::GetHTTPHeader(char *Name)
{
  return NULL;
}


int WebForm::Fetch(void)
{
  int sock;
  int NRead;
  unsigned long inaddr;
  struct sockaddr_in ad;
  struct hostent *hp;
  char *CompleteURL;
  char *POSTData;
  struct WebFormParam *Param;
  int i, j;

  if (Host == NULL || Port == 0 || URL == NULL)
    return -1;

  /* Resolve host */
  memset (&ad, 0, sizeof (ad));
  ad.sin_family = AF_INET;

  inaddr = inet_addr (Host);
  if (inaddr != INADDR_NONE)
    memcpy (&ad.sin_addr, &inaddr, sizeof (inaddr));
  else
    {
      hp = gethostbyname (Host);
      if (hp == NULL)
        {
          perror ("gethostbyname");
          return -1;
        }
      memcpy (&ad.sin_addr, hp->h_addr, hp->h_length);
    }
  ad.sin_port = htons (Port);

  sock = socket (AF_INET, SOCK_STREAM, 0);
  if (sock < 0) {
    perror ("socket");
    return -1;
  }

  if (connect (sock, (struct sockaddr *)&ad, sizeof (ad)) < 0) {
    perror ("connect");
    return -1;
  }

  CompleteURL = (char *)malloc(10240);
  if (!CompleteURL) {
    close(sock);
    return -1;
  }

  if (POSTParams) {
    POSTData = (char *)malloc(10240);
    if (!POSTData) {
      free(CompleteURL);
      close(sock);
      return -1;
    }
  } else
   POSTData = 0;

  HTMLPage = (char *)malloc(10240);
  if (!HTMLPage) {
    if (POSTData)
      free(POSTData);
    free(CompleteURL);
    close(sock);
    return -1;
  }

  strcpy(CompleteURL, URL);
  if (URLParams) {
   strcat(CompleteURL, "?");
   Param = URLParams;
   while (Param) {
      strcat(CompleteURL, Param->Name);
      strcat(CompleteURL, "=");
  for (i=0,j=strlen((char*)CompleteURL);i<strlen((char*)Param->Value);i++) {
      if (Param->Value[i]==' ')
          CompleteURL[j++] = '+'; else
      if ((unsigned char)Param->Value[i]<0x80 && Param->Value[i]>0x20 &&
           Param->Value[i] != '?' && Param->Value[i] != '&' && Param->Value[i] != '=')
          CompleteURL[j++] = Param->Value[i]; else
       {
         CompleteURL[j++]='%';
#define TOX(x)  ((x)>9?(x)+'A'-0xA:(x)+'0')
         CompleteURL[j++]=TOX(Param->Value[i]>>4);
         CompleteURL[j++]=TOX(Param->Value[i]&0xF);
#undef TOX
       }
  }
  if (Param->Next)
    CompleteURL[j++] = '&';
  CompleteURL[j] = 0;
    Param = Param->Next;
   }
  }

  if (POSTParams) {
   POSTData[0] = 0;
   Param = POSTParams;
   while (Param) {
      strcat(POSTData, Param->Name);
      strcat(POSTData, "=");
  for (i=0,j=strlen((char*)POSTData);i<strlen((char *)Param->Value);i++) {
      if (Param->Value[i]==' ')
          POSTData[j++] = '+'; else
      if ((unsigned char)Param->Value[i]<0x80 && Param->Value[i]>0x20 &&
           Param->Value[i] != '?' && Param->Value[i] != '&' && Param->Value[i] != '=')
          POSTData[j++] = Param->Value[i]; else
       {
         POSTData[j++]='%';
#define TOX(x)  ((x)>9?(x)+'A'-0xA:(x)+'0')
         POSTData[j++]=TOX(Param->Value[i]>>4);
         POSTData[j++]=TOX(Param->Value[i]&0xF);
#undef TOX
       }
  }
  if (Param->Next)
    POSTData[j++] = '&';
  POSTData[j] = 0;
    Param = Param->Next;
   }
  }


 if (Debug)
   printf("[WebForm::Fetch] CompleteURL=%s\n",CompleteURL);

 if (Debug && POSTParams) {
   printf("[WebForm::Fetch] POSTData=%s\n", POSTData);
   printf("[WebForm::Fetch] POSTDataLen=%d\n", strlen(POSTData));
  }

 sockprintf(sock, "%s %s HTTP/1.0\n", POSTParams?"POST":"GET", CompleteURL);
 sockprintf(sock, "Host: %s\n", Host);
 if (Referer)
   sockprintf(sock, "Referer: %s\n", Referer);
 if (Cookie)
   sockprintf(sock, "Cookie: %s\n", Cookie);
 if (UserAgent)
   sockprintf(sock, "User-Agent: %s\n", UserAgent);
 if (POSTParams) {
   sockprintf(sock, "Content-length: %d\n", strlen(POSTData));
   sockprintf(sock, "Content-type: application/x-www-form-urlencoded\n");
 }

 if (Debug) {
   printf("[WebForm::Fetch] User-Agent: %s\n", UserAgent);
   printf("[WebForm::Fetch] Referer: %s\n", Referer);
   printf("[WebForm::Fetch] Cookie: %s\n", Cookie);
   printf("[WebForm::Fetch] Host: %s\n", Host);
 }

 sockprintf(sock, "\n");
 if (POSTParams) {
   write(sock, POSTData, strlen(POSTData));
 }

 free(CompleteURL);
 if (POSTData)
   free(POSTData);

 if (Debug)
   printf("[WebForm::Fetch] HTTP/1.0 request sent.\n");

 HTMLPageLen = 0;
 while ((NRead = read (sock, HTMLPage + HTMLPageLen, 10240 - HTMLPageLen)) > 0)
          HTMLPageLen += NRead;
 close (sock);

 if (Debug)
   write(1, HTMLPage, HTMLPageLen);

return 1;
}

int WebForm::FindText(char *Text)
{
  int i;

   for (i = 0; i < HTMLPageLen - strlen(Text); i++)
      if (HTMLPage[i] == Text[0])
            if (!memcmp (HTMLPage + i, Text, strlen(Text)))
              return 1;
  return 0;
}

char *WebForm::FindValue(char *LeadText, char TrailChar)
{
  int i,j;
  char *buf;
  int len;

  len = strlen(LeadText);
  buf = NULL;

      for (i = 0; i < HTMLPageLen - len; i++)
        {
          if (HTMLPage[i] == LeadText[0])
            if (!memcmp (HTMLPage + i, LeadText, len))
              {
                i += len;
                for (j = i; j < HTMLPageLen && HTMLPage[j] != TrailChar; j++);
                if (j == HTMLPageLen)
                  return NULL;
                len = j-i; 
                buf = (char*)malloc(len+1);
                if (buf == NULL)
                    return NULL;
                memcpy(buf, HTMLPage + i, len);
                buf[len] = '\0';
              }
        }
  return buf;
}
